#include "Dialectric.h"

glm::vec3 Dialectric::refract(glm::vec3 uv, glm::vec3 n, float refRat) {
	uv = glm::normalize(uv);
	n = glm::normalize(n);
	float cos_theta = glm::dot(-uv, n);// std::min(glm::dot(-uv, n), 1.0f);
	glm::vec3 r_out_perp = refRat * (uv + cos_theta * n);
	glm::vec3 r_out_parallel = glm::vec3(-sqrt(fabs(1.0f - powf(r_out_perp.length(), 2.0f)))) * n;
	return r_out_perp + r_out_parallel;
}

float Dialectric::schlick(float cosi, float refIx) {
	float r0 = (1 - refIx) / (1 + refIx);
	r0 = r0 * r0;
	return r0 + (1 - r0) * pow((1 - cosi), 5);
}

ray Dialectric::scatter(ray incident, glm::vec3 norm, glm::vec3 pt) {
	norm = glm::normalize(norm);
	glm::vec3 iDir = glm::normalize(incident.getDir());
	
	bool frontFace = (glm::dot(iDir, norm)) < 0.0f;
	float refractRatio = ior;
	if (frontFace) {
		refractRatio = 1.0f / ior;
	}
	else
		norm = -norm;

	float cosTheta = glm::dot(-iDir, norm);// std::min(glm::dot(-iDir, norm), 1.0f);
	float sinTheta = sqrt(1.0f - (cosTheta * cosTheta));

	bool cannotRefract = (refractRatio * sinTheta) > 1.0f;
	glm::vec3 dir;

	float randomNum = ((float)rand() / RAND_MAX);

	if (cannotRefract || schlick(cosTheta, refractRatio) > randomNum)
		dir = (iDir - (norm * 2.0f * (glm::dot(iDir, norm))));
	else
		dir = refract(iDir, norm, refractRatio);

	dir = glm::normalize(dir);

	ray retRay = ray(pt, dir);
	return retRay;
}