#define _CRTDBG_MAP_ALLOC
#include <stdlib.h>
//#include <crtdbg.h>
#include <limits>
#include <algorithm>

#include "Obj.h"
#include "glm/gtx/string_cast.hpp"

#include <iostream>
#include <assert.h>

using namespace std;

float Obj::hit(ray& r) {
	float t = numeric_limits<float>::max();

	glm::vec4 rayDir = glm::vec4(r.getDir(), 0.0f);
	glm::vec4 rayPos = glm::vec4(r.getPos(), 1.0f);
	if (hasTransforms) {
		rayPos = mInv * rayPos;
		rayDir = mInv * rayDir;
	}

	float invDx = 1.0f / rayDir.x;
	float invDy = 1.0f / rayDir.y;
	float invDz = 1.0f / rayDir.z;

	double tx1 = (min.x - rayPos.x) * invDx;
	double tx2 = (max.x - rayPos.x) * invDx;

	double tmin = std::min(tx1, tx2);
	double tmax = std::max(tx1, tx2);

	double ty1 = (min.y - rayPos.y) * invDy;
	double ty2 = (max.y - rayPos.y) * invDy;

	tmin = std::max(tmin, std::min(ty1, ty2));
	tmax = std::min(tmax, std::max(ty1, ty2));

	double tz1 = (min.z - rayPos.z) * invDz;
	double tz2 = (max.z - rayPos.z) * invDz;

	tmin = std::max(tmin, std::min(tz1, tz2));
	tmax = std::min(tmax, std::max(tz1, tz2));

	if (tmin > tmax || tmin < 0)
		return t;

	//do traingle hit here
	t = triangleHit(r);
	return t;
}

float Obj::triangleHit(ray& r) {
	glm::vec3 ptA, ptB, ptC;
	glm::vec3 n;
	glm::vec4 eVec = glm::vec4(r.getPos(), 1.0f);
	glm::vec4 eDir = glm::vec4(r.getDir(), 0.0f);

	if (hasTransforms) {
		eVec = mInv * eVec;
		eDir = mInv * eDir;
	}

	//float A[3][3];
	float a, b, c, d, e, f, g, h, i, j, k, l;
	g = eDir.x;
	h = eDir.y;
	i = eDir.z;

	float minT = numeric_limits<float>::max();
	float t, gamma, beta, M;
	float ei_hf, gf_di, dh_eg, ak_jb, jc_al, bl_kc;

	float aTVal, bTVal, cTVal;
	glm::vec3 testPtA, testPtB, testPtC;
	float aDist, bDist, cDist;
	for (int ix = 0; ix < eleBuf.size(); ix += 3) {
		ptA = glm::vec3(posBuf.at(eleBuf.at(ix) * 3), posBuf.at(eleBuf.at(ix) * 3 + 1), posBuf.at(eleBuf.at(ix) * 3 + 2));
		ptB = glm::vec3(posBuf.at(eleBuf.at(ix + 1) * 3), posBuf.at(eleBuf.at(ix + 1) * 3 + 1), posBuf.at(eleBuf.at(ix + 1) * 3 + 2));
		ptC = glm::vec3(posBuf.at(eleBuf.at(ix + 2) * 3), posBuf.at(eleBuf.at(ix + 2) * 3 + 1), posBuf.at(eleBuf.at(ix + 2) * 3 + 2));

		/*aTVal = glm::dot((ptA - glm::vec3(eVec)), glm::vec3(eDir));
		testPtA = glm::vec3(eVec) + aTVal * glm::vec3(eDir);
		aDist = glm::length(ptA - testPtA);
		bTVal = glm::dot((ptB - glm::vec3(eVec)), glm::vec3(eDir));
		testPtB = glm::vec3(eVec) + bTVal * glm::vec3(eDir);
		bDist = glm::length(ptB - testPtB);
		cTVal = glm::dot((ptC - glm::vec3(eVec)), glm::vec3(eDir));
		testPtC = glm::vec3(eVec) + cTVal * glm::vec3(eDir);
		cDist = glm::length(ptC - testPtC);*/


		aDist = glm::length(glm::cross(glm::vec3(eDir), ptA - glm::vec3(eVec)));
		bDist = glm::length(glm::cross(glm::vec3(eDir), ptB - glm::vec3(eVec)));
		cDist = glm::length(glm::cross(glm::vec3(eDir), ptC - glm::vec3(eVec)));

		//if all 3 points are too far away from the ray, continue
		if (aDist > maxTriDist && bDist > maxTriDist && cDist > maxTriDist)
			continue;


		glm::vec3 e1 = ptB - ptA;
		glm::vec3 e2 = ptC - ptA;

		/*float nX = (e1.y * e2.z) - (e1.z * e2.y);
		float nY = (e1.z * e2.x) - (e1.x * e2.z);
		float nZ = (e1.x * e2.y) - (e1.y * e2.x);
		n = glm::vec3(nX, nY, nZ);*/
		if (hasTransforms) {
			glm::vec3 tmpA = glm::vec3(mInv * glm::vec4(ptA, 1.0f));
			glm::vec3 tmpB = glm::vec3(mInv * glm::vec4(ptB, 1.0f));
			glm::vec3 tmpC = glm::vec3(mInv * glm::vec4(ptC, 1.0f));
			e1 = tmpB - tmpA;
			e2 = tmpC - tmpA;
		}
		n = glm::cross(e1, e2);

		a = ptA.x - ptB.x;
		b = ptA.y - ptB.y;
		c = ptA.z - ptB.z;

		d = ptA.x - ptC.x;
		e = ptA.y - ptC.y;
		f = ptA.z - ptC.z;

		j = ptA.x - eVec.x;
		k = ptA.y - eVec.y;
		l = ptA.z - eVec.z;

		//calc M
		ei_hf = (e * i) - (h * f);
		gf_di = (g * f) - (d * i);
		dh_eg = (d * h) - (e * g);
		M = (a * ei_hf) + (b * gf_di) + (c * dh_eg);

		//calc t
		ak_jb = (a * k) - (j * b);
		jc_al = (j * c) - (a * l);
		bl_kc = (b * l) - (k * c);
		t = (-((f * ak_jb) + (e * jc_al) + (d * bl_kc))) / M;

		if (t < 0 || t > minT) {	//keep track of closest T, compare against it
			//cout << "nope" << endl;
			//return numeric_limits<float>::max();
			continue;
		}
		//calc gamma
		gamma = ((i * ak_jb) + (h * jc_al) + (g * bl_kc)) / M;
		if (gamma < 0 || gamma > 1) {
			//cout << "eh maybe" << endl;
			//return numeric_limits<float>::max();
			continue;
		}

		//calc beta
		beta = ((j * ei_hf) + (k * gf_di) + (l * dh_eg)) / M;
		if (beta < 0 || beta >(1 - gamma)) {
			//cout << "almost" << endl;
			//return numeric_limits<float>::max();
			continue;
		}
		minT = t;
		tempNorm = n;
		//cout << t << endl << M << endl << endl;
		//return t;
	}
	//return abs(posBuf[0]);
	return minT;
}

glm::vec3 Obj::getNorm(glm::vec3& pt) {
	//cout << tempNorm << endl;
	return tempNorm;
}

glm::vec3 Obj::getNormSp(ray& r) {
	//running hit sets tempNorm
	float tVal = hit(r);
	if (tVal < numeric_limits<float>::max())
		return tempNorm;
	return glm::vec3(0);
}

/* copy the data from the shape to this object */
void Obj::createShape(tinyobj::shape_t& shape)
{
	posBuf = shape.mesh.positions;
	norBuf = shape.mesh.normals;
	eleBuf = shape.mesh.indices;
	//init();
}

void Obj::measure() {
	float minX, minY, minZ;
	float maxX, maxY, maxZ;
	float scaleX, scaleY, scaleZ;
	float shiftX, shiftY, shiftZ;
	float epsilon = 0.001f;

	float tMinX, tMinY, tMinZ;
	float tMaxX, tMaxY, tMaxZ;

	minX = minY = minZ = 1.1754E+38F;
	maxX = maxY = maxZ = -1.1754E+38F;

	// Go through all vertices to determine min and max of each dimension
	for (size_t v = 0; v < posBuf.size() / 3; v++) {
		if (posBuf[3 * v + 0] < minX) minX = posBuf[3 * v + 0];
		if (posBuf[3 * v + 0] > maxX) maxX = posBuf[3 * v + 0];

		if (posBuf[3 * v + 1] < minY) minY = posBuf[3 * v + 1];
		if (posBuf[3 * v + 1] > maxY) maxY = posBuf[3 * v + 1];

		if (posBuf[3 * v + 2] < minZ) minZ = posBuf[3 * v + 2];
		if (posBuf[3 * v + 2] > maxZ) maxZ = posBuf[3 * v + 2];
	}



	// From min and max compute necessary scale and shift for each dimension
	float maxExtent, xExtent, yExtent, zExtent;
	xExtent = maxX - minX;
	yExtent = maxY - minY;
	zExtent = maxZ - minZ;
	if (xExtent >= yExtent && xExtent >= zExtent)
	{
		maxExtent = xExtent;
	}
	if (yExtent >= xExtent && yExtent >= zExtent)
	{
		maxExtent = yExtent;
	}
	if (zExtent >= xExtent && zExtent >= yExtent)
	{
		maxExtent = zExtent;
	}
	scaleX = 2.0f / maxExtent;
	shiftX = minX + (xExtent / 2.0f);
	scaleY = 2.0f / maxExtent;
	shiftY = minY + (yExtent / 2.0f);
	scaleZ = 2.0f / maxExtent;
	shiftZ = minZ + (zExtent / 2.0f);

	// Go through all verticies shift and scale them
	for (size_t v = 0; v < posBuf.size() / 3; v++)
	{
		posBuf[3 * v + 0] = (posBuf[3 * v + 0] - shiftX) * scaleX;
		assert(posBuf[3 * v + 0] >= -1.0f - epsilon);
		assert(posBuf[3 * v + 0] <= 1.0f + epsilon);
		posBuf[3 * v + 1] = (posBuf[3 * v + 1] - shiftY) * scaleY;
		assert(posBuf[3 * v + 1] >= -1.0f - epsilon);
		assert(posBuf[3 * v + 1] <= 1.0f + epsilon);
		posBuf[3 * v + 2] = (posBuf[3 * v + 2] - shiftZ) * scaleZ;
		assert(posBuf[3 * v + 2] >= -1.0f - epsilon);
		assert(posBuf[3 * v + 2] <= 1.0f + epsilon);
	}

	minX = (minX - shiftX) * scaleX;
	assert(minX >= -1.0f - epsilon);
	assert(minX <= 1.0f + epsilon);
	minY = (minY - shiftY) * scaleY;
	assert(minY >= -1.0f - epsilon);
	assert(minY <= 1.0f + epsilon);
	minZ = (minZ - shiftZ) * scaleZ;
	assert(minZ >= -1.0f - epsilon);
	assert(minZ <= 1.0f + epsilon);

	maxX = (maxX - shiftX) * scaleX;
	assert(maxX >= -1.0f - epsilon);
	assert(maxX <= 1.0f + epsilon);
	maxY = (maxY - shiftY) * scaleY;
	assert(maxY >= -1.0f - epsilon);
	assert(maxY <= 1.0f + epsilon);
	maxZ = (maxZ - shiftZ) * scaleZ;
	assert(maxZ >= -1.0f - epsilon);
	assert(maxZ <= 1.0f + epsilon);

	min = glm::vec3(minX, minY, minZ);
	max = glm::vec3(maxX, maxY, maxZ);

	/*cout << "ayo we measured" << endl;
	cout << "min: " << glm::to_string(min) << endl;
	cout << "max: " << glm::to_string(max) << endl;*/

	//calculate bounding box of largest triangle in mesh
	tMinX = tMinY = tMinZ = -1.1754E+38F;
	tMaxX = tMaxY = tMaxZ = 1.1754E+38F;
	float triArea = 0.0f;
	float maxTriArea = 0.0f;
	glm::vec3 v0, v1, v2;
	glm::vec3 edge1, edge2, edge3;
	float xLen, yLen, zLen;
	xLen = yLen = zLen = 0.0f;
	maxTriDist = 0.0f;
	for (int i = 0; i < eleBuf.size(); i += 3) {
		v0 = glm::vec3(posBuf.at(eleBuf.at(i) * 3), posBuf.at(eleBuf.at(i) * 3 + 1), posBuf.at(eleBuf.at(i) * 3 + 2));
		v1 = glm::vec3(posBuf.at(eleBuf.at(i + 1) * 3), posBuf.at(eleBuf.at(i + 1) * 3 + 1), posBuf.at(eleBuf.at(i + 1) * 3 + 2));
		v2 = glm::vec3(posBuf.at(eleBuf.at(i + 2) * 3), posBuf.at(eleBuf.at(i + 2) * 3 + 1), posBuf.at(eleBuf.at(i + 2) * 3 + 2));

		edge1 = (v1 - v0);
		edge2 = (v2 - v0);
		edge3 = (v2 - v1);

		xLen = glm::length(edge1);
		if (xLen > maxTriDist)
			maxTriDist = xLen;
		xLen = glm::length(edge2);
		if (xLen > maxTriDist)
			maxTriDist = xLen;
		xLen = glm::length(edge3);
		if (xLen > maxTriDist)
			maxTriDist = xLen;
	}
	maxTriDist *= 0.5f;
	//maxTriBounds = glm::vec3(xLen, yLen, zLen);
}

void Obj::init()
{
	// Send the normal array to the GPU
	if (norBuf.empty()) {
		// Initialize norbuf to zeroes for each vector
		for (int i = 0; i < posBuf.size(); i++) {
			norBuf.push_back(0.0);
		}

		for (int i = 0; i < eleBuf.size(); i += 3) {
			glm::vec3 v0 = glm::vec3(posBuf.at(eleBuf.at(i) * 3), posBuf.at(eleBuf.at(i) * 3 + 1), posBuf.at(eleBuf.at(i) * 3 + 2));
			glm::vec3 v1 = glm::vec3(posBuf.at(eleBuf.at(i + 1) * 3), posBuf.at(eleBuf.at(i + 1) * 3 + 1), posBuf.at(eleBuf.at(i + 1) * 3 + 2));
			glm::vec3 v2 = glm::vec3(posBuf.at(eleBuf.at(i + 2) * 3), posBuf.at(eleBuf.at(i + 2) * 3 + 1), posBuf.at(eleBuf.at(i + 2) * 3 + 2));

			v0 = glm::normalize(v0);
			v1 = glm::normalize(v1);
			v2 = glm::normalize(v2);

			glm::vec3 u = v1 - v0;
			glm::vec3 v = v2 - v0;

			glm::vec3 nf = glm::cross(u, v);//u.cross(v);

			norBuf.at(eleBuf.at(i) * 3) += nf.x;
			norBuf.at(eleBuf.at(i) * 3 + 1) += nf.y;
			norBuf.at(eleBuf.at(i) * 3 + 2) += nf.z;

			norBuf.at(eleBuf.at(i + 1) * 3) += nf.x;
			norBuf.at(eleBuf.at(i + 1) * 3 + 1) += nf.y;
			norBuf.at(eleBuf.at(i + 1) * 3 + 2) += nf.z;

			norBuf.at(eleBuf.at(i + 2) * 3) += nf.x;
			norBuf.at(eleBuf.at(i + 2) * 3 + 1) += nf.y;
			norBuf.at(eleBuf.at(i + 2) * 3 + 2) += nf.z;
		}
		//normalizes the normal vectors
		for (int i = 0; i < norBuf.size() / 3; i += 3) {
			glm::vec3 vert = glm::vec3(norBuf.at(i), norBuf.at(i + 1), norBuf.at(i + 2));
			vert = glm::normalize(vert);
			norBuf.at(i) = vert.x;
			norBuf.at(i + 1) = vert.y;
			norBuf.at(i + 2) = vert.z;
		}
		cout << "normal: " << glm::to_string(glm::vec3(norBuf.at(0), norBuf.at(1), norBuf.at(2))) << endl;
	}
}