#ifndef OBJ_H
#define OBJ_H

#include <string>
#include <vector>
#include <memory>
#include "tiny_obj_loader.h"

#include "Shape.h"

class Obj : public Shape
{
public:
	Obj(tinyobj::shape_t& shape) : Shape{} {
		min = glm::vec3(0);
		max = glm::vec3(0);
		createShape(shape);
		measure();
		init();
		baseColor = glm::vec3(0.6f, 0.6f, 0.9f);
		tempNorm = glm::vec3(0, 0, 1);
		mat = shared_ptr<Lambert>();
		hasTransforms = false;
	}
	Obj(tinyobj::shape_t& shape, shared_ptr<Material> m) : Shape{} {
		min = glm::vec3(0);
		max = glm::vec3(0);
		createShape(shape);
		measure();
		init();
		baseColor = glm::vec3(0.6f, 0.6f, 0.9f);
		tempNorm = glm::vec3(0, 0, 1);
		mat = m;
		hasTransforms = false;
	}
	Obj(tinyobj::shape_t& shape, shared_ptr<Material> m, shared_ptr<glm::mat4> transforms) : Shape{} {
		min = glm::vec3(0);
		max = glm::vec3(0);
		createShape(shape);
		measure();
		init();
		baseColor = glm::vec3(0.6f, 0.6f, 0.9f);
		tempNorm = glm::vec3(0, 0, 1);
		mat = m;
		ctm = *transforms.get();
		mInv = glm::inverse(ctm);
		hasTransforms = true;
	}
	virtual float hit(ray& r);
	float triangleHit(ray& r);
	virtual glm::vec3 getNorm(glm::vec3& pt);
	glm::vec3 getNormSp(ray& r);
	virtual string getType() {
		return type;
	}
	void createShape(tinyobj::shape_t& shape);
	void init();
	void measure();
	vector<float> getPosBuf() { return posBuf; };
	glm::vec3 min;
	glm::vec3 max;
private:
	string type = "obj";
	glm::vec3 tempNorm;
	std::vector<unsigned int> eleBuf;
	std::vector<float> posBuf;
	std::vector<float> norBuf;
	glm::mat4 ctm;
	glm::mat4 mInv;
	bool hasTransforms = false;

	glm::vec3 maxTriBounds;
	float maxTriDist;
};

#endif // !OBJ_H