#include "camera.h"
#include "Shape.h"
#include "Sphere.h"
#include "Plane.h"
#include "Material.h"
#include "Lambert.h"
#include "Specular.h"
#include "Emissive.h"
#include "Dialectric.h"

#include <vector>
#include <sstream>

string removeChars(string s) {
	s.erase(remove(s.begin(), s.end(), '{'), s.end());
	s.erase(remove(s.begin(), s.end(), '<'), s.end());
	s.erase(remove(s.begin(), s.end(), ','), s.end());
	s.erase(remove(s.begin(), s.end(), '>'), s.end());
	s.erase(remove(s.begin(), s.end(), '}'), s.end());
	return s;
}

void setCam(camera& cam, string s) {
	glm::vec3 camPos, camUp, camRight, lookAt;

	stringstream ss;
	ss.str(s);
	string token;
	float x, y, z;
	while (ss >> token) {
		ss >> x >> y >> z;
		if (token == "location")
			camPos = glm::vec3(x, y, z);
		else if (token == "up")
			camUp = glm::vec3(x, y, z);
		else if (token == "right")
			camRight = glm::vec3(x, y, z);
		else if (token == "look_at")
			lookAt = glm::vec3(x, y, z);
	}
	cam = camera(camPos, camRight, camUp, lookAt);
}

void setLight(vector<shared_ptr<Shape>>& objVec, Light& l, string s) {
	stringstream ss;
	ss.str(s);
	string token;
	if (s.find("area_light") == std::string::npos) {
		float x, y, z;
		ss >> x >> y >> z;
		glm::vec3 pos = glm::vec3(x, y, z);
		glm::vec3 rgb = glm::vec3(0);
		while (ss >> token) {
			if (token == "color" && (ss >> token && token == "rgb")) {
				ss >> x >> y >> z;
				rgb = glm::vec3(x, y, z);
			}
		}
		Light light = Light(pos, rgb);
		l = light;
	}
	else {
		float x, y, z;
		float h, w;
		ss >> x >> y >> z;
		glm::vec3 pos = glm::vec3(x, y, z);
		glm::vec3 rgb = glm::vec3(0);
		while (ss >> token) {
			if (token == "color" && (ss >> token && token == "rgb")) {
				ss >> x >> y >> z;
				rgb = glm::vec3(x, y, z);
			}
			else if (token == "area_light") {
				glm::vec3 dims;
				ss >> x >> y >> z;
				dims = glm::vec3(x, y, z);
				ss >> x >> y >> z;
				dims += glm::vec3(x, y, z);
				shared_ptr<Material> m = make_shared<Emissive>(rgb);
				shared_ptr<Shape> r = make_shared<Rectangle>(pos, rgb, dims, m);
				objVec.emplace_back(r);
			}
		}
	}
}

void setSphere(vector<shared_ptr<Shape>>& objVec, string s) {
	//cout << "sphere: " << endl << s << endl;

	stringstream ss;
	ss.str(s);
	string token;
	float x, y, z, r;
	ss >> x >> y >> z >> r;
	glm::vec3 pos = glm::vec3(x, y, z);
	glm::vec3 rgb = glm::vec3(0);
	float reflAmt = 0.0f;
	shared_ptr<Material> m;
	bool setMat = true;
	bool setTex = false;
	bool setBump = false;
	std::vector<string> imgStrs;
	string texString;
	string normString;
	std::vector<glm::mat4> trans;
	while (ss >> token) {
		if (token == "pigment") {
			ss >> token;
			if (token == "color") {
				if (ss >> token && token == "rgb") {
					ss >> x >> y >> z;
					rgb = glm::vec3(x, y, z);
				}
			}
			else if (token == "image_map") {
				ss >> token;	//read in image type
				ss >> token;	//read in image name
				cout << token << endl;
				token.erase(remove(token.begin(), token.end(), '"'), token.end());
				texString = token;
				setTex = true;
			}
		}
		else if (token == "finish") {
			ss >> token;
			if (token == "reflection") {
				ss >> reflAmt;
				m = make_shared<Specular>(rgb, reflAmt);
				setMat = false;
			}
			else if (token == "ior") {
				rgb = glm::vec3(1.0f);
				ss >> reflAmt;
				m = make_shared<Dialectric>(rgb, reflAmt);
				setMat = false;
			}
		}
		else if (token == "scale") {
			ss >> x >> y >> z;
			glm::mat4 sc = glm::scale(glm::mat4(1.0f), glm::vec3(x, y, z));
			trans.push_back(sc);
		}
		else if (token == "rotate") {
			ss >> x >> y >> z;
			glm::vec3 axis;
			float amt;
			if (x != 0) {
				axis = glm::vec3(1, 0, 0);
				amt = x;
			}
			else if (y != 0) {
				axis = glm::vec3(0, 1, 0);
				amt = y;
			}
			else if (z != 0) {
				axis = glm::vec3(0, 0, 1);
				amt = z;
			}
			else
				continue;
			glm::mat4 rot = glm::rotate(glm::mat4(1.0f), glm::radians(amt), axis);
			trans.push_back(rot);
		}
		else if (token == "translate") {
			ss >> x >> y >> z;
			glm::mat4 tr = glm::translate(glm::mat4(1.0f), glm::vec3(x, y, z));
			trans.push_back(tr);
		}
		else if (token == "bump_map") {
			ss >> token;	//read in image type
			ss >> token;	//read in image name
			cout << token << endl;
			token.erase(remove(token.begin(), token.end(), '"'), token.end());
			normString = token;
			setBump = true;
		}
	}
	if(setMat)
		m = make_shared<Lambert>(rgb);
	shared_ptr<Shape> sphere = make_shared<Sphere>(pos, rgb, r, m);

	if (setTex && setBump) {
		imgStrs = { texString, normString };
	}
	else if (setTex && !setBump) {
		imgStrs = { texString };
	}
	else if (!setTex && setBump) {
		imgStrs = { normString };
	}

	if(!trans.empty()){
		glm::mat4 ctm = trans.at(trans.size() - 1);
		for (int i = trans.size() - 2; i > -1; i--)
			ctm *= trans.at(i);
		if (setTex || setBump)
			sphere = make_shared<Sphere>(pos, rgb, r, m, ctm, imgStrs, setTex, setBump);
		else
			sphere = make_shared<Sphere>(pos, rgb, r, m, ctm);
	}
	else if (setTex || setBump) {
		sphere = make_shared<Sphere>(pos, rgb, r, m, imgStrs, setTex, setBump);
	}
	
	objVec.emplace_back(sphere);
}

void objPrep(shared_ptr<glm::mat4>& oTrans, bool& objHasTrans, string s) {
	stringstream ss;
	ss.str(s);
	string token;
	float x, y, z;
	std::vector<glm::mat4> trans;
	while (ss >> token) {
		if (token == "scale") {
			ss >> x >> y >> z;
			glm::mat4 sc = glm::scale(glm::mat4(1.0f), glm::vec3(x, y, z));
			trans.push_back(sc);
		}
		else if (token == "rotate") {
			ss >> x >> y >> z;
			glm::vec3 axis;
			float amt;
			if (x != 0) {
				axis = glm::vec3(1, 0, 0);
				amt = x;
			}
			else if (y != 0) {
				axis = glm::vec3(0, 1, 0);
				amt = y;
			}
			else if (z != 0) {
				axis = glm::vec3(0, 0, 1);
				amt = z;
			}
			else
				continue;
			glm::mat4 rot = glm::rotate(glm::mat4(1.0f), glm::radians(amt), axis);
			trans.push_back(rot);
		}
		else if (token == "translate") {
			ss >> x >> y >> z;
			glm::mat4 tr = glm::translate(glm::mat4(1.0f), glm::vec3(x, y, z));
			trans.push_back(tr);
		}
	}
	if (!trans.empty()) {
		glm::mat4 ctm = trans.at(trans.size() - 1);
		for (int i = trans.size() - 2; i > -1; i--)
			ctm *= trans.at(i);
		oTrans = make_shared<glm::mat4>(ctm);
		objHasTrans = true;
	}
}

void setPlane(vector<shared_ptr<Shape>>& objVec, string s) {
	stringstream ss;
	ss.str(s);
	string token;
	float x, y, z, d;
	ss >> x >> y >> z >> d;
	glm::vec3 pos = glm::vec3(x, y, z);
	glm::vec3 rgb = glm::vec3(0);
	float reflAmt;
	shared_ptr<Material> m;
	bool setMat = true;
	while (ss >> token) {
		if (token == "pigment" && (ss >> token && token == "color" && (ss >> token && token == "rgb"))) {
			ss >> x >> y >> z;
			rgb = glm::vec3(x, y, z);
		}
		else if (token == "finish" && (ss >> token && token == "reflection")) {
			ss >> reflAmt;
			m = make_shared<Specular>(rgb, reflAmt);
			//m = Specular(rgb, reflAmt);
			setMat = false;
		}
	}
	if(setMat)
		m = make_shared<Lambert>(rgb);
	shared_ptr<Shape> plane = make_shared<Plane>(pos, rgb, d, m);
	objVec.emplace_back(plane);
}

void setShape(vector<shared_ptr<Shape>>& objVec, string s, string objT) {
	stringstream ss;
	ss.str(s);
	string token;
	if (objT == "sphere")
		setSphere(objVec, s);
	else if (objT == "plane")
		setPlane(objVec, s);
}

// will need to update when considering non-sphere shapes
void readPov(istream& is, camera& cam, Light& l, vector<shared_ptr<Shape>>& objVec, shared_ptr<glm::mat4>& objTrans, bool& objHasTrans) {

	string s(istreambuf_iterator<char>(is), {});

	int bracketCount = 0;
	stringstream ss;
	ss.str(s);
	string token;
	string objType;
	string collect = "";
	while (ss >> token) {
		if (token == "camera" ||
			token == "sphere" ||
			token == "light_source" ||
			token == "plane" ||
			token == "obj") {
			objType = token;
			ss >> token;
			if (token.find('{') != std::string::npos) {
				bracketCount++;
				collect.append(token + ' ');
				while (bracketCount > 0) {
					ss >> token;
					if (token.find('}') != std::string::npos)
						bracketCount--;
					else if (token.find('{') != std::string::npos)
						bracketCount++;
					collect.append(token + ' ');
				}
			}
			collect = removeChars(collect);
			if (objType == "camera")
				setCam(cam, collect);
			else if (objType == "sphere" ||
				objType == "plane")
				setShape(objVec, collect, objType);
			else if (objType == "light_source")
				setLight(objVec, l, collect);
			else if (objType == "obj")
				objPrep(objTrans, objHasTrans, collect);
			collect = "";
			bracketCount = 0;
		}
	}
}