#pragma once

#ifndef SHAPE_H
#define SHAPE_H

#include <iostream>
#include "Material.h"
#include "Lambert.h"
#include "Specular.h"
#include "Dialectric.h"
#include "ray.h"
//#include "glm::vec3.h"
#include "glm/glm.hpp"
#include "glm/vec3.hpp"
#include "glm/mat4x4.hpp"
#include "glm/gtc/matrix_transform.hpp"

using namespace std;

class Shape
{
public:
	Shape() : baseColor(glm::vec3(0.5f)), pos(glm::vec3(0.0f)), mat(make_shared<Lambert>()) {}
	Shape(glm::vec3 p) : baseColor(glm::vec3(0.5f)), pos(p), mat(make_shared<Lambert>()) {}
	Shape(glm::vec3 p, glm::vec3 rgb) : baseColor(rgb), pos(p), mat(make_shared<Lambert>()) {}
	Shape(glm::vec3 p, glm::vec3 rgb, shared_ptr<Material>& m) : baseColor(rgb), pos(p), mat(m) {}

	virtual float hit(ray& r);
	glm::vec3 getPos() { return pos; }
	virtual glm::vec3 getRGB(glm::vec3& pt) { return baseColor; }
	virtual shared_ptr<Material> getMat() { return mat; }
	virtual glm::vec3 getNorm(glm::vec3& pt);
	virtual glm::vec3 getNormSp(ray& r);
	virtual string getType() { return type; }
	virtual bool hasTex() { return false; }
protected:
	string type = "Shape";
	glm::vec3 pos;
	glm::vec3 baseColor;
	shared_ptr<Material> mat;
};

#endif