#pragma once

#ifndef VEC3_H
#define VEC3_H

#include <cmath>
#include <iostream>

using namespace std;

class vec3 {
public:
	vec3(double in1, double in2, double in3) : val{ in1, in2, in3 } {}
	vec3() : val{ 0,0,0 } {} //default constructor
	vec3(double in) : val{ in,in,in } {} //alt constructor

	//getters
	double x() const { return val[0]; }
	double y() const { return val[1]; }
	double z() const { return val[2]; }
	void setX(double inX) { val[0] = inX; }
	void setY(double inY) { val[1] = inY; }
	void setZ(double inZ) { val[2] = inZ; }

	//overload operators
	vec3& operator+=(const vec3& v) {
		val[0] += v.val[0];
		val[1] += v.val[1];
		val[2] += v.val[2];
		return *this;
	}

	const vec3 operator+(const vec3& v) const {
		vec3 res = *this;
		res += v;
		return res;
	}

	vec3& operator-=(const vec3& v) {
		val[0] -= v.val[0];
		val[1] -= v.val[1];
		val[2] -= v.val[2];
		return *this;
	}

	const vec3 operator-(const vec3& v) const {
		vec3 res = *this;
		res -= v;
		return res;
	}

	//read more about what friend does
	friend ostream& operator<<(ostream& os, const vec3& v) {
		os << "<" << v.x() << ", " << v.y() << ", " << v.z() << ">";
		return os;
	}

	//mult by scalar

	vec3& operator*=(const double Sc) {
		val[0] *= Sc;
		val[1] *= Sc;
		val[2] *= Sc;
		return *this;
	}

	vec3& operator*=(const vec3 v) {
		val[0] *= v.x();
		val[1] *= v.y();
		val[2] *= v.z();
		return *this;
	}

	const vec3 operator*(const vec3& v) const {
		vec3 res = *this;
		res *= v;
		return res;
	}

	vec3& operator/=(const double Sc) {
		val[0] /= Sc;
		val[1] /= Sc;
		val[2] /= Sc;
		return *this;
	}

	vec3& operator/=(const vec3 v) {
		val[0] /= v.x();
		val[1] /= v.y();
		val[2] /= v.z();
		return *this;
	}

	const vec3 operator/(const vec3& v) const {
		vec3 res = *this;
		res /= v;
		return res;
	}

	double leng() const {
		float xsq = powf(val[0], 2.0f);
		float ysq = powf(val[1], 2.0f);
		float zsq = powf(val[2], 2.0f);
		return sqrt(xsq + ysq + zsq);
	}

	void normalize() {
		double l = this->leng();
		val[0] /= l;
		val[1] /= l;
		val[2] /= l;
	}

	vec3 cross(vec3 v) {
		float x = this->y() * v.z() - this->z() * v.y();
		float y = -(this->x() * v.z() - this->z() * v.x());
		float z = this->x() * v.y() - this->y() * v.x();
		return vec3(x, y, z);
	}

	float dot(vec3 v) {
		float res = this->x() * v.x();
		res += (this->y() * v.y());
		res += (this->z() * v.z());
		return res;
	}

public:
	double val[3];
};

//Type aliases for vec3
using point3D = vec3; //3D point
using color = vec3; //RGB color

#endif
